@@ -13,7 +13,11 @@
 extern int check_img_inventory(void);
 extern int write_img_inventory(InventoryEntry *he);
 extern int prepare_inventory(InventoryEntry *he);
-extern int add_post_prepare_cb(int (*actor)(void *data), void *data);
+struct pprep_head {
+	int (*actor)(struct pprep_head *);
+	struct pprep_head *next;
+};
+extern void add_post_prepare_cb(struct pprep_head *);
 extern bool deprecated_ok(char *what);
 extern int cr_dump_tasks(pid_t pid);
 extern int cr_pre_dump_tasks(pid_t pid);
@@ -25,17 +29,16 @@ extern int cr_dedup(void);
 extern int check_add_feature(char *arg);
 extern void pr_check_features(const char *offset, const char *sep, int width);
 
-#define add_post_prepare_cb_once(actor, data)			\
-	({							\
-		static int __cb_called = 0;			\
-		int ret = 0;					\
-								\
-		if (!__cb_called) {				\
-			ret = add_post_prepare_cb(actor, data);	\
-			__cb_called = 1;			\
-		}						\
-								\
-		ret;						\
-	})
+#define PPREP_HEAD_INACTIVE	((struct pprep_head *)-1)
+
+#define add_post_prepare_cb_once(phead) do {		 \
+		if ((phead)->next == PPREP_HEAD_INACTIVE)\
+			add_post_prepare_cb(phead);	 \
+	} while (0)
+
+#define MAKE_PPREP_HEAD(name) struct pprep_head name = {	\
+			.next = PPREP_HEAD_INACTIVE,		\
+			.actor = name##_cb,			\
+	}
 
 #endif /* __CR_CRTOOLS_H__ */
